import * as express from "express";
import * as _ from "lodash";
import * as jwt from "jsonwebtoken";
import * as bcrypt from "bcryptjs";
import {
  KeystrokeDataType,
  SignUpPayload,
  TypedRequest,
  USER_TYPE,
} from "../config/types";
import Auth from "../models/Auth";
import Candidates from "../models/Candidate";
import Organizations from "../models/Organization";
import {
  processKeystrokeData,
  findUser,
  createSignupDataFromProcessedData,
  addAttemptToKeystrokeData,
  computeDataTendencies,
  processAttempt,
} from "../utilities/keystroke";
import { JWT_SECRET } from "../config/contants";
import { authMiddleware } from "../middlewares/auth";

const router = express.Router();

router.post("/signup", async (req: TypedRequest<{}, SignUpPayload>, res) => {
  const {
    passwords,
    keydown,
    keyup,
    email,
    userType,
    candidate,
    organization,
  } = req.body;
  const attemptCount = passwords.length;

  const auth = await Auth.findOne({ email });

  if (auth) {
    return res
      .status(403)
      .json({ success: false, msg: "User already exists in DB" });
  }

  const processedAttempts: KeystrokeDataType[] = Array(attemptCount)
    .fill(0)
    .map((v, i) =>
      processKeystrokeData({ keydown: keydown[i], keyup: keyup[i] })
    );

  const passwordsEqual = processedAttempts.every((v) =>
    _.isEqual(v.full.keys, processedAttempts[0].full.keys)
  );

  if (!passwordsEqual) {
    return res
      .status(403)
      .json({ success: false, msg: "Passwords don't match" });
  }

  try {
    let newUser;
    if (userType === USER_TYPE.CANDIDATE) {
      newUser = await Candidates.create(candidate);
    } else {
      newUser = await Organizations.create(organization);
    }

    const signupData = await createSignupDataFromProcessedData(
      email,
      passwords,
      userType,
      newUser.id,
      processedAttempts
    );

    await Auth.create(signupData);

    return res.json({
      success: true,
    });
  } catch (error) {
    return res
      .status(500)
      .json({ success: false, msg: "Error signing up", error });
  }
});

router.post("/login", async (req, res) => {
  const { password, keydown, keyup, email, controls, userType } = req.body;

  const auth = await Auth.findOne({ email });

  if (!auth) {
    return res
      .status(403)
      .json({ success: false, msg: "User does not exist in DB" });
  }

  const processedAttempt: KeystrokeDataType = processKeystrokeData({
    keydown,
    keyup,
  });

  const isMatch = await bcrypt.compare(password, auth.password);

  const credentialsValid =
    isMatch &&
    _.isEqual(processedAttempt.hold.keys, auth.keystrokeData.hold.keys) &&
    _.isEqual(processedAttempt.flight.keys, auth.keystrokeData.flight.keys) &&
    _.isEqual(processedAttempt.dd.keys, auth.keystrokeData.dd.keys) &&
    _.isEqual(processedAttempt.full.keys, auth.keystrokeData.full.keys);

  if (!credentialsValid) {
    return res.status(403).json({ success: false, msg: "Invalid Credentials" });
  }

  const result = processAttempt({
    userKeystrokeData: auth.keystrokeData,
    attemptKeystrokeData: processedAttempt,
    controls,
  });

  if (result.accepted) {
    const newUserData = addAttemptToKeystrokeData({
      userData: auth,
      attemptKeystrokeData: processedAttempt,
    });

    await Auth.updateOne({ email }, { $set: newUserData });
  }

  let user;

  if (userType === USER_TYPE.CANDIDATE) {
    user = await Candidates.findById(auth.userId);
  } else {
    user = await Organizations.findById(auth.userId);
  }

  const token = await jwt.sign(auth, JWT_SECRET, { expiresIn: "2h" });

  return res.json({
    user,
    result,
    token,
    db: {
      hold: auth.keystrokeData.hold.means,
      flight: auth.keystrokeData.flight.means,
      dd: auth.keystrokeData.dd.means,
      full: auth.keystrokeData.full.means,
    },
    filteredDb: {
      hold: auth.keystrokeData.hold.filteredMeans,
      flight: auth.keystrokeData.flight.filteredMeans,
      dd: auth.keystrokeData.dd.filteredMeans,
      full: auth.keystrokeData.full.filteredMeans,
    },
    attempt: {
      hold: processedAttempt.hold.times,
      flight: processedAttempt.flight.times,
      dd: processedAttempt.dd.times,
      full: processedAttempt.full.times,
    },
  });
});

router.post("/validateKeystroke", authMiddleware, (req, res) => {
  const processedKeystrokeData = processKeystrokeData(req.body);
  res.json(processedKeystrokeData);
});

router.get("/tendencies/:userId", authMiddleware, async (req, res) => {
  findUser(req.params.userId)
    .then((_auth) => {
      if (_auth) {
        return res.json({
          db: _auth.keystrokeData,
          calc: computeDataTendencies(_auth.keystrokeData),
        });
      } else {
        throw new Error("User not found");
      }
    })
    .catch((error) => res.status(500).send(error));
});

module.exports = router;
